/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU Lesser
 * General Public License Version 2.1 only ("LGPL") or the Common Development and
 * Distribution License ("CDDL")(collectively, the "License"). You may not use this
 * file except in compliance with the License. You can obtain a copy of the CDDL at
 * http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
 * http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
 * specific language governing permissions and limitations under the License. When
 * distributing the software, include this License Header Notice in each file and
 * include the full text of the License in the License file as well as the
 * following notice:
 * 
 * NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
 * (CDDL)
 * For Covered Software in this distribution, this License shall be governed by the
 * laws of the State of California (excluding conflict-of-law provisions).
 * Any litigation relating to this License shall be subject to the jurisdiction of
 * the Federal Courts of the Northern District of California and the state courts
 * of the State of California, with venue lying in Santa Clara County, California.
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or only
 * the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
 * include this software in this distribution under the [CDDL or LGPL Version 2.1]
 * license." If you don't indicate a single choice of license, a recipient has the
 * option to distribute your version of this file under either the CDDL or the LGPL
 * Version 2.1, or to extend the choice of license to its licensees as provided
 * above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
 * Version 2 license, then the option applies only if the new code is made subject
 * to such option by the copyright holder. 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "imi_view.h"
#include "imi_view_modern.h"
#include "imi_view_classic.h"

int CIMIView::s_CandiWindowSize = 6;
int CIMIViewFactory::sm_mapInit = 0;

static std::map<TWCHAR, TWCHAR>   halfPunc2Full;
static std::map<TWCHAR, TWCHAR>   halfSimbol2Full;
static std::set<TWCHAR>           termPuncs;
static TWCHAR                     wc_RightDblQuote;

static void                       initSimbolPuncMap(void);

#pragma setlocale("zh_CN.UTF-8")

CIMIView *
CIMIViewFactory::createView(int viewType)
{
    if (sm_mapInit == 0) {
        initSimbolPuncMap();
        sm_mapInit = 1;
    }

    switch (viewType) {
    case SVT_MODERN:
        return new CIMIModernView();
    case SVT_CLASSIC:
        return new CIMIClassicView();
    }
    return NULL;
}

CIMIView::CIMIView()
    : m_pIC(NULL), m_bRightFullDblQuote(false),
      m_CN(true), m_FullPunc(true), m_FullSimbol(false),
      mp_winHandler(NULL), m_pPref(NULL)
{
}

CIMIView::~CIMIView()
{
    if (m_pPref) delete m_pPref;
}

void
CIMIView::setPreference(const CIMIOptions* pp)
{
    if (pp) {
        if (m_pPref) delete m_pPref;
        m_pPref = pp->clone();
    }
}

int
CIMIView::getViewType(void)
    { return CIMIViewFactory::SVT_NONE; }

void
CIMIView::attachIC(CIMIContext* pIC)
    { m_pIC = pIC; }

unsigned
CIMIView::clearIC(void)
    {  m_pIC->clear();  return 0; }

void
CIMIView::updateWindows(unsigned int mask)
    { }

int
CIMIView::onKeyEvent(unsigned keycode, unsigned keyvalue, unsigned modifier)
    { return false; }

void
CIMIView::setStatusAttrValue(int key, int value)
{
    switch(key) {
    case CIMIWinHandler::STATUS_ID_CN:
        m_CN = (value != 0);
        if (mp_winHandler)
            mp_winHandler->updateStatus( key, value );
        break;
    case CIMIWinHandler::STATUS_ID_FULLPUNC:
        m_FullPunc = (value != 0);
        if (mp_winHandler)
            mp_winHandler->updateStatus( key, value );
        break;
    case CIMIWinHandler::STATUS_ID_FULLSIMBOL:
        m_FullSimbol = (value != 0);
        if (mp_winHandler)
            mp_winHandler->updateStatus( key, value );
        break;
    }
}

int
CIMIView::getStatusAttrValue(int key)
{
    switch(key) {
    case CIMIWinHandler::STATUS_ID_CN:
        return (m_CN)?1:0;
    case CIMIWinHandler::STATUS_ID_FULLPUNC:
        return (m_FullPunc)?1:0;
    case CIMIWinHandler::STATUS_ID_FULLSIMBOL:
        return (m_FullSimbol)?1:0;
    }
    return 0;
}

bool
CIMIView::isPuncChar(TWCHAR key)
{
    return (halfPunc2Full.find(key) != halfPunc2Full.end());
}

bool
CIMIView::isTermPuncWide(TWCHAR wc)
{
    return (termPuncs.find(wc) != termPuncs.end());
}

bool
CIMIView::isSimbolChar(TWCHAR key)
{
    return (halfSimbol2Full.find(key) != halfSimbol2Full.end());
}

TWCHAR
CIMIView::getFullPunc(TWCHAR key)
{
    std::map<TWCHAR, TWCHAR>::const_iterator it = halfPunc2Full.find(key);
    if (it != halfPunc2Full.end()) {
        if (key == TWCHAR('"')) {
            m_bRightFullDblQuote = !m_bRightFullDblQuote;
            return (m_bRightFullDblQuote)?(it->second):wc_RightDblQuote;
        } else
            return it->second;
    }
    return key;
}

TWCHAR
CIMIView::getFullSimbol(TWCHAR key)
{
    std::map<TWCHAR, TWCHAR>::const_iterator it = halfSimbol2Full.find(key);
    if (it != halfSimbol2Full.end())
        return it->second;
    return key;
}

//Static functions in this file
static void
initSimbolPuncMap(void)
{
    static const char half_punc[]   = ",;!?.:^\\\"_";
    static const char punc[]        = "，；！？。：…、“—";
    static int        isTermPunc[]  = {1,1,1,1,1,1,1,0,0,0,0};
    static const char rdq[]         = "”";
    TWCHAR full[256];

#ifdef HAVE_ICONV_H
    iconv_t icv = iconv_open(TWCHAR_ICONV_NAME, "UTF-8");
    TIConvSrcPtr src = (TIConvSrcPtr)punc;
    size_t srclen = strlen(punc)+1;
    char *dst = (char*)full;
    size_t dstlen = 256*sizeof(TWCHAR);
    iconv(icv, &src, &srclen, &dst, &dstlen);
#else // !HAVE_ICONV_H
    char *src = (char*)punc;
    memset(&full[0], 0, sizeof(full));
    MBSTOWCS(&full[0], src, (sizeof(full) / sizeof(TWCHAR)) - 1);
#endif // HAVE_ICONV_H

    TWCHAR*wp=full;
    for (src=(char*)half_punc; *src; ++src)
        halfPunc2Full[*src] = *wp++;

    for (int sz=strlen(half_punc); sz > 0; ){
        --sz;
        if (isTermPunc[sz])
            termPuncs.insert(full[sz]);
    }

    static const char half_simbol[] = "azAZ<>()[]{}@#$%&*+|";
    static const char simbol[] = "ａｚＡＺ《》（）〔〕｛｝＠＃￥％＆＊＋｜";

#ifdef HAVE_ICONV_H
    src = (TIConvSrcPtr)simbol;
    srclen = strlen(simbol)+1;
    dst = (char*)full;
    dstlen = 256*sizeof(TWCHAR);
    iconv(icv, &src, &srclen, &dst, &dstlen);
#else // !HAVE_ICONV_H
    src = (char*)simbol;
    memset(&full[0], 0, sizeof(full));
    MBSTOWCS(&full[0], src, (sizeof(full) / sizeof(TWCHAR)) - 1);
#endif // HAVE_ICONV_H

    TWCHAR wch = full[0];
    for (char h=half_simbol[0], t=half_simbol[1]; h <= t; ++h)
        halfSimbol2Full[h] = wch++;

    wch = full[2];
    for (char h=half_simbol[2], t=half_simbol[3]; h <= t; ++h)
        halfSimbol2Full[h] = wch++;

    wp=full+4;
    for (src=(char*)(half_simbol+4); *src; ++src)
        halfPunc2Full[*src] = *wp++;

#ifdef HAVE_ICONV_H
    src = (TIConvSrcPtr)rdq;
    srclen = strlen(rdq)+1;
    dst = (char*)full;
    dstlen = 256*sizeof(TWCHAR);
    iconv(icv, &src, &srclen, &dst, &dstlen);
    iconv_close(icv);
#else // !HAVE_ICONV_H
    src = (char*)rdq;
    memset(&full[0], 0, sizeof(full));
    MBSTOWCS(&full[0], src, (sizeof(full) / sizeof(TWCHAR)) - 1);
#endif // HAVE_ICONV_H
    wc_RightDblQuote = full[0];
}

void
CIMIView::getPreeditString(IPreeditString& ps)
{
    ps.clear();
}

void
CIMIView::getCandidateList(ICandidateList& cl, int start, int size)
{
    cl.clear();
}

void
CIMIView::getInputString(wstring& s)
{
    s.clear();
}

int
CIMIView::onCandidatePageRequest(int pgno, bool relative)
{
    return 0;
}

int
CIMIView::onCandidateSelectRequest(int index)
{
    return 0;
}
