/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU Lesser
 * General Public License Version 2.1 only ("LGPL") or the Common Development and
 * Distribution License ("CDDL")(collectively, the "License"). You may not use this
 * file except in compliance with the License. You can obtain a copy of the CDDL at
 * http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
 * http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
 * specific language governing permissions and limitations under the License. When
 * distributing the software, include this License Header Notice in each file and
 * include the full text of the License in the License file as well as the
 * following notice:
 * 
 * NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
 * (CDDL)
 * For Covered Software in this distribution, this License shall be governed by the
 * laws of the State of California (excluding conflict-of-law provisions).
 * Any litigation relating to this License shall be subject to the jurisdiction of
 * the Federal Courts of the Northern District of California and the state courts
 * of the State of California, with venue lying in Santa Clara County, California.
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or only
 * the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
 * include this software in this distribution under the [CDDL or LGPL Version 2.1]
 * license." If you don't indicate a single choice of license, a recipient has the
 * option to distribute your version of this file under either the CDDL or the LGPL
 * Version 2.1, or to extend the choice of license to its licensees as provided
 * above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
 * Version 2 license, then the option applies only if the new code is made subject
 * to such option by the copyright holder. 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "lattice_states.h"
#include "imi_data.h"

void
CBoneInnerData::print(std::string& prefix)
{
    printf("%s",prefix.c_str());
    printf("[BoneInnerData @%X]", this);
    if (m_BWType != NoBestWordStartHere) {
        printf((m_BWType == UserSelectedBestWord)?"U ":"B ");
        std::string nullstr;
        m_BestWord.print(nullstr);
    }
    printf("\n");
    printf("%s",prefix.c_str());
    printf("[status list @%X]\n", &m_LatticeNodes);

    prefix += "    ";
    CLatticeStates::iterator it = m_LatticeNodes.begin();
    CLatticeStates::iterator itEnd = m_LatticeNodes.end();
    for (; it != itEnd; ++it) {
        it->print(prefix);
        printf("\n");
    }
    prefix.resize(prefix.size() - 4);

    fflush(stdout);
}

void
TLatticeState::print(std::string& prefix)
{
    printf("%s",prefix.c_str());
#ifdef _USE_RAW_PROBABILITY
    char valbuf[256];
    m_Score.toString(valbuf);
    printf("<State(%d:%d), from word %d, score %s>", m_State.getLevel(), m_State.getIdx(), m_BackTraceWordId, valbuf);
#else
    printf("<State(%d:%d), from word %d, score %E>", m_State.getLevel(), m_State.getIdx(), m_BackTraceWordId, m_Score);
#endif

    fflush(stdout);
}

const unsigned CLatticeStates::beam_width;

/**
 * clear all existing states, and assure the vector's size.
 * Of cause the reserve space action only useful right after construction,
 * but just let it be.
 */
void
CLatticeStates::clear()
{
    m_Vec.clear();
    m_Heap.clear();
    m_VecIdxInHeap.clear();
    m_Map.clear();

    m_Vec.reserve(beam_width);
    m_VecIdxInHeap.reserve(beam_width);
    m_Heap.reserve(beam_width);
}

/**
 * Following conditions may be encountered:
 *    - node's StateKey is alredy in the map
 *        -# But its score is larger than the original one: \n
 *                   ===> do nothing.
 *        -# its score is less than the original one: \n
 *                   ===> replace original node with new one, ironDown it
 *    - node is a new node
 *        -# Vector is full
 *            - and node's score is >= the largest score of all existing node:\n
 *                   ===> do nothing.
 *            - node score < the largest score of all existing node:\n
 *                   ===> Relpace the largest node in the vector with new node,
 *                   ironDown it (0 on the heap)
 *        -# Vector is not full: \n
 *             ===> push_back the new node in vector, build heap node in the
 *             heap tail. bubbleUp it.
 */
void
CLatticeStates::push_back(const TLatticeState& node)
{
    std::map<CStateKey, int>::iterator itMap;

    itMap = m_Map.find(node.m_State);
    if (itMap != m_Map.end()) {
        TLatticeState& oldNode = m_Vec[itMap->second];
        if (node.m_Score < oldNode.m_Score) {
            oldNode = node;
            ironDown(m_VecIdxInHeap[itMap->second]);
        }
    } else {
        if (m_Vec.size() >= beam_width) {
            TLatticeState& oldNode = m_Vec[m_Heap[0]];
            if (node.m_Score < oldNode.m_Score) {
                itMap = m_Map.find(oldNode.m_State);
                m_Map.erase(itMap);
                m_Map[node.m_State] = m_Heap[0];
                oldNode = node;
                ironDown(0);
            }
        } else {
            m_Map[node.m_State] = m_Vec.size();
            m_VecIdxInHeap.push_back(m_Vec.size());
            m_Vec.push_back(node);
            m_Heap.push_back(m_Heap.size());
            bubbleUp(m_Heap.size()-1);
        }
    }
}

/**
 * The index of heap's corresponding node like a bubble,
 * it should go up in the heap until it goes to the top
 * or meet lighter bubble.
 */
void
CLatticeStates::bubbleUp(int idxInHeap)
{
    while (idxInHeap > 0) {
        int idxNode = m_Heap[idxInHeap];
        int parentInHeap = (idxInHeap-1) / 2;
        int parentNode = m_Heap[parentInHeap];

        // if child's score > parent's, try to swap them
        if (m_Vec[parentNode].m_Score < m_Vec[idxNode].m_Score) {
            m_VecIdxInHeap[idxNode] = parentInHeap;
            m_VecIdxInHeap[parentNode]  = idxInHeap;
            m_Heap[idxInHeap] = parentNode;
            m_Heap[parentInHeap] = idxNode;
            idxInHeap = parentInHeap;
        } else {
            break;
        }
    }
}

/**
 * The index of heap's corresponding node like a iron,
 * it should go down in the heap, until to the bottom, or
 * meet heavier bubble.
 */
void
CLatticeStates::ironDown(int idxInHeap)
{
    int sz = m_Heap.size(), lcInHeap, rcInHeap;
    while ((lcInHeap = 2*idxInHeap + 1) < sz) {
        int idxNode = m_Heap[idxInHeap];
        int lcNode = m_Heap[lcInHeap];

        // make the rcInHeap become the child idx in heap whose node's
        // score is the largest of the two child node
        rcInHeap = lcInHeap + 1;
        if (rcInHeap >= sz ||
            m_Vec[m_Heap[rcInHeap]].m_Score <= m_Vec[lcNode].m_Score)
            rcInHeap = lcInHeap;
        int rcNode = m_Heap[rcInHeap];

        // if parent score not the largest, try to swap it with better child
        if (m_Vec[idxNode].m_Score < m_Vec[rcNode].m_Score) {
            m_VecIdxInHeap[idxNode] = rcInHeap;
            m_VecIdxInHeap[rcNode]  = idxInHeap;
            m_Heap[idxInHeap] = rcNode;
            m_Heap[rcInHeap] = idxNode;
            idxInHeap = rcInHeap;
        } else {
            break;
        }
    }
}
